package de.mmt.mpeg.framework.data;

/**
 * This represents a 16x16 macro block. A macro block always uses YCbCr mode with a reduced
 * resolution of the Cb and Cr layers.
 * 
 * @author ruempel
 * @author MaNi
 */

public class MacroBlock
{
public final static int BLOCKSIZE = 16;
public final static int XFACTOR = 2;
public final static int YFACTOR = 2;
private int[][] yMatrix = new int[BLOCKSIZE][BLOCKSIZE];
private int[][] crMatrix = new int[BLOCKSIZE/XFACTOR][BLOCKSIZE/YFACTOR];
private int[][] cbMatrix = new int[BLOCKSIZE/XFACTOR][BLOCKSIZE/YFACTOR];
private boolean isDiff = false;
private int offsetX = 0;
private int offsetY = 0;

/**
 * Creates a new, empty macro block.
 */

public MacroBlock() {}
    
/**
 * Duplicates a macro block.
 * 
 * @param mb the original macro block
 */

public MacroBlock(MacroBlock mb)
  {
   multiArrayCopy(mb.yMatrix,this.yMatrix);
   multiArrayCopy(mb.cbMatrix,this.cbMatrix);
   multiArrayCopy(mb.crMatrix,this.crMatrix);
   this.isDiff = mb.isDiff;
   this.offsetX = mb.offsetX;
   this.offsetY = mb.offsetY;
  }
    
/**
 * Indicates whether this is a difference-coded macro block.
 * 
 * @return true if it is difference-coded, false otherwise
 */

public boolean isDifferenceCoded()
  {
   return isDiff;
  }

/**
 * Set whether this is a difference-coded macro block.
 * 
 * @param isDiff true if it is difference-coded, false otherwise
 */

public void setDifferenceCoded(boolean isDiff)
  {
   this.isDiff = isDiff;
  }
   
/**
 * Returns where (horizontally) the most similar block in the previous frame lies (relative to this block).
 * Only applicable if this is a difference-coded macro block.
 * 
 * @return the horizontal offset of the most similar block in the previous frame, e.g., -2
 */

public int getOffsetX()
  {
   return offsetX;
  }

/**
 * Store the horizontal position (relative to this block) of a similar block in the previous frame. 
 * Only applicable if this is a difference-coded macro block.
 * 
 * @param offsetX the horizontal offset of a similar block in the previous frame
 */

public void setOffsetX(int offsetX)
  {
   this.offsetX = offsetX;
  }

/**
 * Returns where (vertically) the most similar block in the previous frame lies (relative to this block).
 * Only applicable if this is a difference-coded macro block.
 * 
 * @return the vertical offset of the most similar block in the previous frame, e.g., 1
 */

public int getOffsetY()
  {
   return offsetY;
  }

/**
 * Store the vertical position (relative to this block) of a similar block in the previous frame. 
 * Only applicable if this is a difference-coded macro block.
 * 
 * @param offsetX the vertical offset of a similar block in the previous frame
 */

public void setOffsetY(int offsetY)
  {
   this.offsetY = offsetY;
  }

/**
 * Returns the Y value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * @return the Y value of the pixel at the given position
 */

public int getY(int x,int y)
  {
   return yMatrix[x][y];
  }

/**
 * Returns the Cb value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * @return the Cb value of the pixel at the given position
 */

public int getCb(int x,int y)
  {
   return cbMatrix[x][y];
  }

/**
 * Returns the Cr value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * @return the Cr value of the pixel at the given position
 */

public int getCr(int x,int y)
  {
   return crMatrix[x][y];
  }

/**
 * Sets the Y value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * param value the new Y value of the pixel at the given position
 */

public void setY(int x,int y,int value)
  {
   yMatrix[x][y] = value;
  }

/**
 * Sets the Cb value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * param value the new Cb value of the pixel at the given position
 */

public void setCb(int x,int y,int value)
  {
   cbMatrix[x][y] = value;
  }

/**
 * Sets the Cr value of the pixel at the given position within this macro block.
 * 
 * @param x the horizontal position
 * @param y the vertical position
 * param value the new Cr value of the pixel at the given position
 */

public void setCr(int x,int y,int value)
  {
   crMatrix[x][y] = value;
  }

/**
 * Generates an XML representation of this macro block.
 */

public String toString()
  {
   StringBuffer s = new StringBuffer();
   s.append("<macroblock>\n");
     
   if (this.isDiff)
     {
      s.append("<offsetx>" + offsetX + "</offsetx>\n");
      s.append("<offsety>" + offsetY + "</offsety>\n");
     }
     
   s.append("<y>\n");   
   writeMatrix(s,yMatrix);   
   s.append("</y>\n<cb>\n");   
   writeMatrix(s,cbMatrix);
   s.append("</cb>\n<cr>\n");
   writeMatrix(s,crMatrix);
   s.append("</cr>\n</macroblock>\n");
     
   return s.toString();
  }

private static void writeMatrix(StringBuffer s,int[][] matrix)
  {
   for (int i = 0; i < matrix.length; i++)
     {
      int[] column = matrix[i];
     
      for (int j = 0; j < column.length; j++)
        {
         s.append(matrix[i][j]);
       
         if (i+1 < matrix.length || j+1 < column.length) s.append(" ");
        }
     } 
  }
    
/**
 * Helper function for deep copying of multidimensional arrays
 * 
 * @param source
 * @param destination
 */

protected static void multiArrayCopy(int[][] source,int[][] destination)
  {
   for (int a=0;a<source.length;a++)
     {
      System.arraycopy(source[a],0,destination[a],0,source[a].length);
     }
  }
}
