package de.mmt.mpeg.framework.data;

import java.util.ArrayList;
import java.util.List;


/**
 * A MatrixTriple is a generic container for a set of three two-dimensional matrices.
 * It can thus be used to store either RGB or YCC images.
 * 
 * @author ruempel
 * @author MaNi
 */

public class MatrixTriple
{
private List<List<Integer>> m1;
private List<List<Integer>> m2;
private List<List<Integer>> m3;
private int xFactor = 1;
private int yFactor = 1;

public final static int R_LAYER = 0;
public final static int G_LAYER = 1;
public final static int B_LAYER = 2;
public final static int Y_LAYER = 0;
public final static int Cb_LAYER = 1;
public final static int Cr_LAYER = 2;

/**
 * Create a new MatrixTriple with three empty matrices.
 */

public MatrixTriple()
  {
   this.m1 = new ArrayList<List<Integer>>();
   this.m2 = new ArrayList<List<Integer>>();
   this.m3 = new ArrayList<List<Integer>>();
  }


/**
 * Create a new MatrixTriple with a given resolution, setting all individual pixel values to 0.
 * All matrices will share the same resolution.
 * 
 * @param xRes the horizontal resolution in pixels, e.g., 320
 * @param yRes the vertical resolution in pixels, e.g., 240
 */

public MatrixTriple(int xRes,int yRes)
  {
   this(xRes,yRes,1,1);
  }

/**
 * Create a new MatrixTriple with a given resolution, setting all individual pixel values to 0.
 * The resolution of the second and third matrix can be reduced by supplying division factors.
 * For example, using a resolution of 320x240 and factors 2 and 2, the resulting matrix triple
 * will be initialized with M1 of 320x240 values and M2 and M3 with 160x120 values.
 * 
 * Warning: Using resolutions that can not be divided by the respective factors will result in
 * errors.
 * 
 * @param xRes the horizontal resolution in pixels, e.g., 320
 * @param yRes the vertical resolution in pixels, e.g., 240
 * @param xFactor the factor by which the horizontal resolution of matrix #2 and #3 is divided, e.g., 2
 * @param yFactor the factor by which the vertical resolution of matrix #2 and #3 is divided, e.g., 2
 */

public MatrixTriple(int xRes,int yRes,int xFactor,int yFactor)
  {
   this();
   this.xFactor = xFactor;
   this.yFactor = yFactor;
       
   for (int i = 0; i < xRes; i++)
     {
      ArrayList<Integer> row = new ArrayList<Integer>(yRes);
          
      for (int j = 0; j < yRes; j++)
        {
         row.add(0);
        }
          
      this.m1.add(row);
     }
       
   for (int i = 0; i < xRes/xFactor; i++)
     {
      ArrayList<Integer> rowCb = new ArrayList<Integer>(yRes/yFactor);
      ArrayList<Integer> rowCr = new ArrayList<Integer>(yRes/yFactor);
         
      for (int j = 0; j < yRes/yFactor; j++)
        {
         rowCb.add(0);
         rowCr.add(0);
        }

      this.m2.add(rowCb);
      this.m3.add(rowCr);
     }
  }

/**
 * Returns matrix #1. In the context of the exercise, this will be either the R layer of an RGB
 * matrix triple or the Y layer of an YCbCr matrix triple.
 * 
 * @return the first matrix, e.g., the R or Y layer
 */

public List<List<Integer>> getM1()
  {
   return m1;
  }

/**
 * Sets matrix #1. In the context of the exercise, this will be either the R layer of an RGB
 * matrix triple or the Y layer of an YCbCr matrix triple.
 * 
 * @param matrix the new matrix for the first layer
 */

public void setM1(List<List<Integer>> matrix)
  {
   this.m1 = matrix;
  }

/**
 * Returns matrix #2. In the context of the exercise, this will be either the G layer of an RGB
 * matrix triple or the Cb layer of an YCbCr matrix triple.
 * 
 * @return the second matrix, e.g., the G or Cb layer
 */

public List<List<Integer>> getM2()
  {
   return m2;
  }

/**
 * Sets matrix #2. In the context of the exercise, this will be either the G layer of an RGB
 * matrix triple or the Cb layer of an YCbCr matrix triple.
 * 
 * @param matrix the new matrix for the second layer
 */

public void setM2(List<List<Integer>> matrix)
  {
   this.m2 = matrix;
  }

/**
 * Returns matrix #3. In the context of the exercise, this will be either the B layer of an RGB
 * matrix triple or the Cr layer of an YCbCr matrix triple.
 * 
 * @return the third matrix, e.g., the B or Cr layer
 */

public List<List<Integer>> getM3()
  {
   return m3;
  }

/**
 * Sets matrix #3. In the context of the exercise, this will be either the B layer of an RGB
 * matrix triple or the Cr layer of an YCbCr matrix triple.
 * 
 * @param matrix the new matrix for the third layer
 */

public void setM3(List<List<Integer>> matrix)
  {
   this.m3 = matrix;
  }

/**
 * Returns a vector of the three values at a given position. For example, in a RGB matrix triple,
 * if the pixel at the given position is red, you will receive {255,0,0}.
 * This method respects the x and y factors of this matrix triple, so you will receive the appropriate
 * values if these layers have a reduced resolution.
 * 
 * @param y the y position of the pixel to read
 * @param x the x position of the pixel to read
 * @return the values at the given position as a vector
 */

public int[] getPixelAsVector(int y,int x)
  {
   int[] vector = new int[] {0,0,0};
   vector[0] = m1.get(x).get(y);
   vector[1] = m2.get(x/this.xFactor).get(y/this.yFactor);
   vector[2] = m3.get(x/this.xFactor).get(y/this.yFactor);
   
   return vector;
  }

/**
 * Returns a matrix of the three values at a given position. For example, in a RGB matrix triple,
 * if the pixel at the given position is red, you will receive {{255},{0},{0}}.
 * This method respects the x and y factors of this matrix triple, so you will receive the appropriate
 * values if these layers have a reduced resolution.
 * 
 * @param y the y position of the pixel to read
 * @param x the x position of the pixel to read
 * @return the values at the given position as a matrix
 */

public int[][] getPixelAsMatrix(int y,int x)
  {
   int[] vector = this.getPixelAsVector(y,x);
   return new int[][]{{vector[0]},{vector[1]},{vector[2]}};
  }

/**
 * Returns the value of a single layer of a pixel at the given position.
 * The layer has to be addressed by an identifier, e.g., MatrixTriple.R_LAYER.
 * 
 * @param y the y position of the pixel to read
 * @param x the x position of the pixel to read
 * @param layer the layer of the pixel to read, e.g., MatrixTriple.Cb_LAYER
 * @return the value of the given layer at the given position, e.g., 255
 */

public int getPixelValue(int y,int x,int layer)
  {
   switch(layer)
     {
      case R_LAYER:
         return m1.get(x).get(y);
      case G_LAYER:
         return m2.get(x/this.xFactor).get(y/this.yFactor);
      case B_LAYER:
         return m3.get(x/this.xFactor).get(y/this.yFactor);
     }
   
   throw new IllegalArgumentException(layer+" is an invalid layer identifier.");
  }

/**
 * Returns the horizontal resolution of the matrix triple. This is not affected by
 * the second or third matrix having a reduced resolution. 
 * 
 * @return the horizontal resolution of this matrix triple
 */

public int getWidth()
  {
   return m1.size();
  }

/**
 * Returns the vertical resolution of the matrix triple. This is not affected by
 * the second or third matrix having a reduced resolution. 
 * 
 * @return the vertical resolution of this matrix triple
 */

public int getHeight()
  {
   return m1.get(0).size();
  }

public String toString()
  {
   return "M1: " + m1.toString() + "\nM2: " + m2.toString() + "\nM3: " + m3.toString();
  }
}