package de.mmt.mpeg.framework.util;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.ParsingException;
import nu.xom.ValidityException;
import de.mmt.mpeg.framework.data.MacroBlock;
import de.mmt.mpeg.framework.data.MatrixTriple;

/**
 * Contains various utility methods.
 * 
 * @author mani
 */

public class Util
{
/**
 * Exports a macro block frame to a file.
 * 
 * @param frame the frame to export
 * @param path the path under which to store the frame 
 */
public static void exportBlockMatrix(MacroBlock[][] frame,String path) throws IOException
  {
   BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(path),"UTF8"));
   out.write("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n" + "<frame>\n<width>" + frame.length
           + "</width>\n<height>" + frame[0].length + "</height>\n");
     
   for (int i = 0; i < frame.length; i++)
     {
      for (int j = 0; j < frame[0].length; j++)
        {
         out.write(frame[i][j].toString());
        }
     }
     
   out.write("</frame>");
   out.close();
  }

/**
 * Transforms an image with given filename to rgb matrices.
 * 
 * @param filename the filename of the image to transform
 * @return an RGB matrix triple representation of the image
 */

public static MatrixTriple getRGBMatricesFromFilename(String filename) throws IOException
  {
   InputStream is = Util.class.getResourceAsStream(filename);
   BufferedImage img = ImageIO.read(is);
   
   return getRGBMatricesFromImage(img);
  }

/**
 * Transforms an image with given filename to rgb matrices.
 * 
 * @param img the image object to transform
 * @return an RGB matrix triple representation of the image
 */

public static MatrixTriple getRGBMatricesFromImage(BufferedImage img)
  {
    System.out.println("Width: " + img.getData().getWidth() + " Height: "
        + img.getData().getHeight());

    MatrixTriple rgbMatrices = new MatrixTriple();
    List<List<Integer>> redPixels = rgbMatrices.getM1();
    List<List<Integer>> greenPixels = rgbMatrices.getM2();
    List<List<Integer>> bluePixels = rgbMatrices.getM3();

    // write rgb info of each pixel in matrix
    for(int i = 0; i < img.getWidth(); i++) {
        List<Integer> redPixelRow = new ArrayList<Integer>();
        List<Integer> greenPixelRow = new ArrayList<Integer>();
        List<Integer> bluePixelRow = new ArrayList<Integer>();

        redPixels.add(redPixelRow);
        greenPixels.add(greenPixelRow);
        bluePixels.add(bluePixelRow);

        for(int j = 0; j < img.getHeight(); j++) {
            int rgbValue = img.getRGB(i,j);
            redPixelRow.add((rgbValue >> 16) & 0xff);
            greenPixelRow.add((rgbValue >> 8) & 0xff);
            bluePixelRow.add((rgbValue) & 0xff);
        }
    }

    return rgbMatrices;
  }

/**
 * Transforms an image from YCC matrix representation to RGB.
 * 
 * @param ycc the matrix to transform
 * @return the same matrix in RGB representation
 */

public static MatrixTriple transformYCC2RGB(MatrixTriple ycc)
  {
   MatrixTriple rgb = new MatrixTriple();
   float[][] trafo = {{1f, 0f, 1.402f},{1f,-0.344136f,-0.714136f},{1f,1.772f,0f}};

    // perform matrix multiplication and set values for rgb
    for (int column = 0; column < ycc.getWidth(); column++)
      {
       List<Integer> rRow = new ArrayList<Integer>();
       List<Integer> gRow = new ArrayList<Integer>();
       List<Integer> bRow = new ArrayList<Integer>();
       rgb.getM1().add(rRow);
       rgb.getM2().add(gRow);
       rgb.getM3().add(bRow);
       
       for (int row = 0; row < ycc.getHeight(); row++)
         {           
          int[] yccPixel = ycc.getPixelAsVector(row,column);
          int[][] yccVector = {{yccPixel[0]},{yccPixel[1]-128},{yccPixel[2]-128}};
          float[][] rgbVector = multiply(trafo,yccVector);
           
          // cut off values outside of scale
          if (rgbVector[0][0] > 255) rgbVector[0][0] = 255;
          if (rgbVector[1][0] > 255) rgbVector[1][0] = 255;
          if (rgbVector[2][0] > 255) rgbVector[2][0] = 255;
          if (rgbVector[0][0] < 0)   rgbVector[0][0] = 0;
          if (rgbVector[1][0] < 0)   rgbVector[1][0] = 0;
          if (rgbVector[2][0] < 0)   rgbVector[2][0] = 0;

          // round to integer
          rRow.add(Math.round(rgbVector[0][0]));
          gRow.add(Math.round(rgbVector[1][0]));
          bRow.add(Math.round(rgbVector[2][0]));
         }
      }
    
   return rgb;
  }

/**
 * Converts a frame into a matrix triple
 * 
 * @param frame the frame to convert
 * @param doPixeling whether every macroblock as a whole should be represented by only one color value
 * @return the resulting matrix triple
 */

public static MatrixTriple convertToMatrixTriple(MacroBlock[][] frame,boolean doPixeling)
  {
   int MBSIZE = MacroBlock.BLOCKSIZE;
   int width = frame.length;
   int height = frame[0].length;
   MatrixTriple ycc = new MatrixTriple(width*MacroBlock.BLOCKSIZE,height*MacroBlock.BLOCKSIZE,1,1);
   
   for (int x = 0; x < width; x++)
     {    
      for (int y = 0; y < height; y++)
        {
         MacroBlock block = frame[x][y];
                        
         // do pixeling if desired
         if (doPixeling)
           { 
            int ySum = 0;
            int cbSum = 0;
            int crSum = 0;
            
            // calculate average color
            for (int i = 0; i < MBSIZE; i++)
              {
               for (int j = 0; j < MBSIZE; j++)
                 {
                  ySum += block.getY(i,j);
                  
                  // lower resolution
                  if (i < MBSIZE/MacroBlock.XFACTOR && j < MBSIZE/MacroBlock.YFACTOR)
                    {
                     cbSum += block.getCb(i,j);
                     crSum += block.getCr(i,j);
                    }
                 }
              }
            
            int yAvg = ySum/(MBSIZE*MBSIZE);
            int cbAvg = cbSum/(MBSIZE/MacroBlock.XFACTOR*MBSIZE/MacroBlock.XFACTOR);
            int crAvg = crSum/(MBSIZE/MacroBlock.XFACTOR*MBSIZE/MacroBlock.XFACTOR);
            
            // set the same for all pixels of the macro block
            for (int i = 0; i < MBSIZE; i++)
              {
               for (int j = 0; j < MBSIZE; j++)
                 {
                  ycc.getM1().get(x*MBSIZE+i).set(y*MBSIZE+j,yAvg);
                  ycc.getM2().get(x*MBSIZE+i).set(y*MBSIZE+j,cbAvg);
                  ycc.getM3().get(x*MBSIZE+i).set(y*MBSIZE+j,crAvg);
                 }
              } 
           }
         else
           {
            // set all pixels from the macro block
            for (int i = 0; i < MBSIZE; i++)
              {
               for (int j = 0; j < MBSIZE; j++)
                 {
                  if (x*MBSIZE+i >= ycc.getWidth() || y*MBSIZE+j >= ycc.getHeight())
                    {
                     throw new RuntimeException("Error during conversion: Matrix dimensions too small for macroblock frame.");
                    }
                  else
                    {
                     ycc.getM1().get(x*MBSIZE+i).set(y*MBSIZE+j,block.getY(i,j));
                     ycc.getM2().get(x*MBSIZE+i).set(y*MBSIZE+j,block.getCb(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR));
                     ycc.getM3().get(x*MBSIZE+i).set(y*MBSIZE+j,block.getCr(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR));
                    }
                 }
              } 
           }
        }
     }
   
   return ycc;
  }

/**
 * Restores an original frame from a difference-coded frame and the other frame used to encode the difference frame.
 * @param diffFrame the difference-coded frame to decode
 * @param previousFrame the "previous" frame used to encode the diffFrame
 * @return the original frame
 */

public static MatrixTriple restoreOriginalFrame(MacroBlock[][] diffFrame,MacroBlock[][] previousFrame)
  {
   int MBSIZE = MacroBlock.BLOCKSIZE;
   int width = diffFrame.length;
   int height = diffFrame[0].length;
   MatrixTriple ycc = new MatrixTriple(width*MacroBlock.BLOCKSIZE,height*MacroBlock.BLOCKSIZE,1,1);
   
   // re-create frame2 pixels     
   for (int x = 0; x < diffFrame.length; x++)
     {
      for (int y = 0; y < diffFrame[x].length; y++)
        {
         MacroBlock block = diffFrame[x][y];
                  
         for (int i = 0; i < MBSIZE; i++)
           {
            for (int j = 0; j < MBSIZE; j++)
              {
               int pixelY  = block.getY(i,j);
               int pixelCb = diffFrame[x][y].getCb(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR);
               int pixelCr = diffFrame[x][y].getCr(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR);
               
               if (block.isDifferenceCoded())
                 {
                  if (previousFrame == null) throw new IllegalArgumentException("Macro block #"+x+","+y+" refers to previous frame, but that frame is null.");
                  
                  pixelY  += previousFrame[x+diffFrame[x][y].getOffsetX()][y+diffFrame[x][y].getOffsetY()].getY(i,j);
                  pixelCb += previousFrame[x+diffFrame[x][y].getOffsetX()][y+diffFrame[x][y].getOffsetY()].getCb(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR);
                  pixelCr += previousFrame[x+diffFrame[x][y].getOffsetX()][y+diffFrame[x][y].getOffsetY()].getCr(i/MacroBlock.XFACTOR,j/MacroBlock.YFACTOR);

                  if (pixelY < 0 || pixelY > 255)
                    {
                     System.err.println("Y value '"+pixelY+"' out of range, cutting off...");       
                     pixelY = Math.min(Math.max(pixelY,100),0);
                    }
                  if (pixelCb < 0 || pixelCb > 255)
                    {
                     System.err.println("Cb value '"+pixelCb+"' out of range, cutting off...");       
                     pixelCb = Math.min(Math.max(pixelCb,100),0);
                    }
                  if (pixelCr < 0 || pixelCr > 255)
                    {
                     System.err.println("Cr value '"+pixelCr+"' out of range, cutting off...");       
                     pixelCr = Math.min(Math.max(pixelCr,100),0);
                    }
                 }

               ycc.getM1().get(x*MBSIZE+i).set(y*MBSIZE+j,pixelY);
               ycc.getM2().get(x*MBSIZE+i).set(y*MBSIZE+j,pixelCb);
               ycc.getM3().get(x*MBSIZE+i).set(y*MBSIZE+j,pixelCr);
              }
           }
        }
     }
   
   return ycc;
  }

/**
 * Takes a matrix triple of reduced Cb and Cr resolution and interpolates full Y-resolution 
 * for these color layers. Works only if Y resolution is an integer multiple of Cb and Cr resolution and
 * if Cb and Cr share the same resolution. 
 * 
 * @param yccMatrices the color matrix to upscale
 * @return a color matrix with Cb and Cr having the same resolution as Y
 */

public static MatrixTriple upscaleColorResolution(MatrixTriple yccMatrices)
  {
   int yResX  = yccMatrices.getM1().size();
   int cbResX = yccMatrices.getM2().size();
   int crResX = yccMatrices.getM3().size();

   // check for empty matrices
   if (yResX == 0) throw new IllegalArgumentException("Y-Ebene ist leer!");
   if (cbResX == 0) throw new IllegalArgumentException("Cb-Ebene ist leer!");
   if (crResX == 0) throw new IllegalArgumentException("Cr-Ebene ist leer!");
   
   int yResY  = yccMatrices.getM1().get(0).size();
   int cbResY = yccMatrices.getM2().get(0).size();
   int crResY = yccMatrices.getM3().get(0).size();
   
   // already full resolution
   if (yResY == cbResY && yResY == crResY && yResX == cbResX && yResX == crResX) return yccMatrices;
   
   // resolution mismatch
   if (cbResY > yResY) throw new IllegalArgumentException("Cb-Ebene hat eine hhere Y-Auflsung als Y-Ebene!");
   if (crResY > yResY) throw new IllegalArgumentException("Cr-Ebene hat eine hhere Y-Auflsung als Y-Ebene!");
   if (cbResX > yResX) throw new IllegalArgumentException("Cb-Ebene hat eine hhere X-Auflsung als Y-Ebene!");
   if (crResX > yResX) throw new IllegalArgumentException("Cr-Ebene hat eine hhere X-Auflsung als Y-Ebene!");
   if (cbResY != crResY) throw new IllegalArgumentException("Cb- und Cr-Ebenen haben unterschiedliche Y-Auflsungen!");
   if (cbResX != crResX) throw new IllegalArgumentException("Cb- und Cr-Ebenen haben unterschiedliche X-Auflsungen!");

   int yFactor = yResY/cbResY;
   int xFactor = yResX/cbResX;
   
   if (yFactor*cbResY != yResY) throw new IllegalArgumentException("Die Y-Auflsung der Y-Ebene ist kein ganzzahliges Vielfaches der Cb- und Cr-Ebenen!");
   if (xFactor*cbResX != yResX) throw new IllegalArgumentException("Die X-Auflsung der Y-Ebene ist kein ganzzahliges Vielfaches der Cb- und Cr-Ebenen!");

   // create new full resolution matrices
   List<List<Integer>> cbMatrix = new ArrayList<List<Integer>>(yResX);
   List<List<Integer>> crMatrix = new ArrayList<List<Integer>>(yResX);
   
   for (int x = 0; x < yResX; x++)
     {
      List<Integer> cbRow = new ArrayList<Integer>(yResY);
      List<Integer> crRow = new ArrayList<Integer>(yResY);
      
      for (int y = 0; y < yResY; y++)
        {
         // use value repetition to fill full resolution matrices
         cbRow.add(yccMatrices.getM2().get(x/xFactor).get(y/yFactor));
         crRow.add(yccMatrices.getM3().get(x/xFactor).get(y/yFactor));
        }
      
      cbMatrix.add(cbRow);
      crMatrix.add(crRow);
     }
   
   // replace lowres color layers with upscaled ones
   yccMatrices.setM2(cbMatrix);
   yccMatrices.setM3(crMatrix);
   
   return yccMatrices;
  }

/**
 * Matrix multiplication.
 * 
 * @param a matrix
 * @param b matrix
 * @return the multiplied matrix
 */

public static float[][] multiply(float a[][],int b[][])
  {
   int aRows = a.length,aColumns = a[0].length,bRows = b.length,bColumns = b[0].length;

   if (aColumns != bRows) throw new IllegalArgumentException("A:Rows: "+aColumns+" did not match B:Columns "+bRows+".");

   float[][] result = new float[aRows][bColumns];

   for (int i = 0; i < aRows; i++)
     {
      for (int j = 0; j < bColumns; j++)
        {
         for (int k = 0; k < aColumns; k++)
           {
            result[i][j] += a[i][k]*b[k][j];
           }
        }
     }

   return result;
  }

/**
 * Creates a displayable image object from a matrix triple
 * 
 * @param triple the matrix triple to convert
 * @return an image object
 */

public static BufferedImage createImageFromMatrixTriple(MatrixTriple triple)
  {   
   // create canvas image
   BufferedImage restoredImg = new BufferedImage(triple.getWidth(),triple.getHeight(),BufferedImage.TYPE_INT_RGB);

   for (int y = 0; y < triple.getHeight(); y++)
     {
      for (int x = 0; x < triple.getWidth(); x++)
        {
         Color c = new Color(triple.getM1().get(x).get(y),triple.getM2().get(x).get(y),triple.getM3().get(x).get(y));
         restoredImg.setRGB(x,y,c.getRGB());
        }
     }
   
   return restoredImg;
  }


/**
 * Imports a macro block frame from a file.
 * 
 * @param filename the file to read
 * 
 * @return a macro block frame
 */

public static MacroBlock[][] importBlockMatrix(String filename) throws FileNotFoundException
  {
   InputStream is = new FileInputStream(new File(filename));

   // parse xml
   Builder parser = new Builder(false);
   Document doc = null;
   
   try
     {
      doc = parser.build(is);
     }
   catch (ValidityException e)
     {
      System.out.println("Die Makroblockdatei '"+filename+"' enthlt ungltiges XML: "+e.getMessage());
     }
   catch (ParsingException e)
     {
      System.out.println("Die Makroblockdatei '"+filename+"' konnte nicht verarbeitet werden: "+e.getMessage());
     }
   catch (IOException e)
     {
      System.out.println("Unbekannter Fehler beim Einlesen der Makroblockdatei '"+filename+"': "+e.getMessage());
     }
   
   Element root = doc.getRootElement();
   int frameWidth = Integer.valueOf(root.getChildElements("width").get(0).getValue().trim());
   int frameHeight = Integer.valueOf(root.getChildElements("height").get(0).getValue().trim());
   Elements macroBlocks = root.getChildElements("macroblock");
   
   MacroBlock[][] frame = new MacroBlock[frameWidth][frameHeight];

   for (int i = 0; i < macroBlocks.size(); i++)
     {
      Element macroBlockElement = macroBlocks.get(i);
      MacroBlock block = new MacroBlock();
      
      if (macroBlockElement.getChildElements("offsetx").size() > 0)
        {
         int offsetX = Integer.valueOf(macroBlockElement.getChildElements("offsetx").get(0).getValue().trim());
         int offsetY = Integer.valueOf(macroBlockElement.getChildElements("offsety").get(0).getValue().trim());
         
         block.setOffsetX(offsetX);
         block.setOffsetY(offsetY);
         block.setDifferenceCoded(true);
        }

      // fill mb
      String valueList = macroBlockElement.getChildElements("y").get(0).getValue().trim();
      String[] valueArray = valueList.split(" ");
      
      for (int j = 0; j < valueArray.length; j++)
        {
         int value = Integer.valueOf(valueArray[j]);
         block.setY(j/MacroBlock.BLOCKSIZE,j%MacroBlock.BLOCKSIZE,value);
        }

      valueList = macroBlockElement.getChildElements("cb").get(0).getValue().trim();
      valueArray = valueList.split(" ");
      
      for (int j = 0; j < valueArray.length; j++)
        {
         int value = Integer.valueOf(valueArray[j]);
         block.setCb(j/(MacroBlock.BLOCKSIZE/MacroBlock.YFACTOR),j%(MacroBlock.BLOCKSIZE/MacroBlock.YFACTOR),value);
        }

      valueList = macroBlockElement.getChildElements("cr").get(0).getValue().trim();
      valueArray = valueList.split(" ");
      
      for (int j = 0; j < valueArray.length; j++)
        {
         int value = Integer.valueOf(valueArray[j]);
         block.setCr(j/(MacroBlock.BLOCKSIZE/MacroBlock.YFACTOR),j%(MacroBlock.BLOCKSIZE/MacroBlock.YFACTOR),value);
        }
      
      frame[i/frameHeight][i%frameHeight] = block;
     }

   return frame;
  }
}